#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bitmap.h"
#include "color.h"

/**********************************************************************************************
 * This method will create a blank BitMap structure from headers that are passed in.
 *
 * @param  struct BitMapHeader *hdr This is the header that is to be used for this bitmap.
 * @param  struct BitMapInfo* bmi This is the bitmap information that is to eb used.
 * @return A blank bitmap will be pointed to by the return value.
 **********************************************************************************************/
struct BitMap *createBlankBitmapFromHeaders(struct BitMapHeader *hdr, struct BitMapInfo* bmi)
{
  struct BitMap* retValue = createBlankBitmap(bmi->Width, bmi->Height);
  (void)memcpy(&retValue->header, hdr, sizeof(struct BitMapHeader));
  (void)memcpy (&retValue->info, bmi, sizeof(struct BitMapInfo));
  return retValue;
}

/**********************************************************************************************
 * This method will create a blank BitMap structure from the image height and width.
 *
 * @param  int width This is the width of the image.
 * @param  int height This is the height of the image.
 * @return A blank bitmap will be pointed to by the return value.
 **********************************************************************************************/
struct BitMap *createBlankBitmap(int width, int height)
{
   int i, j;
   struct BitMap *returnValue = (struct BitMap *)malloc(sizeof(struct BitMap));
   
   BitMapHeader* bmh = createBitmapHeader(width, height);
   BitMapInfo *bmi = createBitMapInfoStructure( width,  height);
   
   memcpy(&returnValue->header, bmh, sizeof(BitMapHeader));
   memcpy(&returnValue->info, bmi, sizeof(BitMapInfo));

   char* address = (char*)&returnValue->header;
   address = (char*)& returnValue->info;
   
   
   returnValue->height = height;
   returnValue->width = width;
   
   /*allocate memory*/
   /* Start by allocating a large block for everything. */
   // Everything includes an array of row pointers, and an array of pixels for the image.
   int mallocSize = returnValue->height * sizeof(struct Pixel *) + (returnValue->width * returnValue->height) * (sizeof(struct Pixel));
   returnValue->rawDataArray = malloc(mallocSize);
   
   for (unsigned int idx = 0; idx < mallocSize; idx++)
   {
     ((unsigned char*)(returnValue->rawDataArray))[idx] = 0;
   }
   
   // Initialize the structures appropriately.
   returnValue->pixelArray = (struct Pixel*)((char*)returnValue->rawDataArray + sizeof(struct Pixel*) * returnValue->height);
   
   for(i=0; i < returnValue->height; i++) {
      for(j=0; j < returnValue->width; j++) {
	  Pixel* p = getPixel(returnValue, j, i);

	  p->Red = 0;
	  p->Green = 0;
	  p->Blue = 0;
      }
   }
   free(bmi);
   free(bmh);
   return returnValue;
} 


/**********************************************************************************************
 * This function will return a pointer to a given pixel.
 *
 * @param  struct BitMap *bmp This is a pointer to the bitmap that is to be used.
 * @param  int x This is the x offset (horizontal) into the image.
 * @param  int y This is the y offset (vertical) into the image.
 * @return A pointer to the pixel will be returned.
 **********************************************************************************************/
Pixel* getPixel(struct BitMap* bmp, int x, int y)
{
 unsigned int offset = x + y*(bmp->width);
  return &(bmp->pixelArray[offset]);
}

/**********************************************************************************************
 * This function will update the internal HSI pixels from the RGB data/.
 *
 * @param  struct BitMap *bmp This is a pointer to the bitmap that is to be used.
 * @return No return from the method.
 **********************************************************************************************/
void updateHSIPixels(struct BitMap* bmp)
{
  int r, c;
  for (r = 0; r < bmp->height; r++)
  {
    for (c = 0; c < bmp->width; c++)
    {
	  Pixel* p = getPixel(bmp, c, r);
	  convertRGBToHSI(p, &p->hsirepresentation);
    }
  }
}

/**********************************************************************************************
 * This function will update the internal RGB pixels from the HSI data/.
 *
 * @param  struct BitMap *bmp This is a pointer to the bitmap that is to be used.
 * @return No return from the method.
 **********************************************************************************************/
void updateRGBPixels(struct BitMap* bmp)
{
  int r, c;
  for (r = 0; r < bmp->height; r++)
  {
    for (c = 0; c < bmp->width; c++)
    {
	  Pixel* p = getPixel(bmp, c, r);
	  convertHSIToRGB(&p->hsirepresentation, p);
    }
  }
}


/**********************************************************************************************
 * This method will destroy a given bitmap, freeing all memory.
 *
 * @param  struct BitMap* This si the bitmap to be freed.
 **********************************************************************************************/
void destroyBitmap(struct BitMap* bmp)
{
  // Start by freeing the main allocated memory.
  free(bmp->rawDataArray);
  
  // Now free the structure itself.
  free(bmp);
}

/**********************************************************************************************
 * This method will read a bitmap from a given file, loading the data into memory.
 *
 * @param  char* filename This si the name of the file to open.
 * @return A bitmap will be pointed to by the return value.
 **********************************************************************************************/
struct BitMap *readBitmapFromFile(char *filename)
{
   FILE *BMP_in = fopen (filename, "rb");
   int i, j;
   int padding;
   
   struct BitMapHeader hdr;
   struct BitMapInfo bmi;
 
   if (filename == NULL) {
      printf("\nCannot open file\n");
      exit(1);
   }
   
   readBitmapHeader(&hdr, BMP_in);
   readBitMapInfoStructure(&bmi, BMP_in);
   
   // Create an empty bitmap structure to hold the data.
   struct BitMap *returnValue = createBlankBitmapFromHeaders(&hdr, &bmi);
   
   // Read the bitmap from the file.
   for(i= returnValue->height-1; i >= 0; i--) {
      padding = 0;
      for(j=0; j < returnValue->width; j++) {
	  Pixel* p = getPixel(returnValue, j, i);

	  fread(&p->Blue, 1, 1,BMP_in);
	  fread(&p->Green, 1, 1,BMP_in);
	  fread(&p->Red, 1, 1,BMP_in);
	  padding+=3;
      }
      if(padding%4 != 0) {
	        char junk[16];
            padding = 4 - (padding%4);
            fread(&junk[0], padding, 1,BMP_in);            
         }
   }
   fclose(BMP_in);
   
   return returnValue;
}

/**********************************************************************************************
 * This method will write a bitmap to a given file, storing the data from memory.
 *
 * @param  char* filename This si the name of the file to write.
 * @param struct BitMap bitmap This si the bitmap to write out.
 **********************************************************************************************/
void writeBitmapToFile(char *filename, struct BitMap *bmp)
{
   int i, j, padding;
   FILE *BMP_out = fopen(filename, "wb");
   if (BMP_out == NULL) {
      printf("\nCannot open file\n");
      exit(1);
   }
   writeBitmapHeader(&bmp->header, BMP_out);
   writeBitMapInfoStructure(&bmp->info, BMP_out);

   for(i=bmp->height-1; i >=0; i--) {
      padding=0;

      for(j=0; j < bmp->width; j++) {
		 Pixel* p = getPixel(bmp, j, i);
         fwrite(&p->Blue, sizeof(unsigned char), 1, BMP_out);
         fwrite(&p->Green, sizeof(unsigned char), 1, BMP_out);
         fwrite(&p->Red, sizeof(unsigned char), 1, BMP_out);
         padding += 3;;
      }

	  if(padding % 4 != 0) {
	        char pads[] = {0,0,0,0};
            padding = 4 - (padding%4);
            fwrite(&pads[0], padding, 1, BMP_out);
        }
   }
   fclose(BMP_out);
}

/**********************************************************************************************
 * This method will read a bitmap header from a given file, loading the data into memory.
 *
 * @param  BitMapHeader* destination This is where the header data is to be placed.
 * @param  FILE* fp This is the name of the file to open.
 **********************************************************************************************/
void readBitmapHeader(BitMapHeader *destination, FILE* fp)
{
  unsigned char buffer[14];

  fread(&buffer, 14, 1, fp);

  destination->Type=((unsigned short)buffer[1])<<8 | buffer[0];
  destination->Size=  ((unsigned int)buffer[5])<<24;
  destination->Size|= ((unsigned int)buffer[4])<<16;
  destination->Size|= ((unsigned int)buffer[3])<< 8;
  destination->Size|= ((unsigned int)buffer[2])<<0;

  destination->Reserved1 = buffer[7]<<8 | buffer[6];
  destination->Reserved2= buffer[9]<<8 | buffer[8];
  destination->Offset=buffer[13]<<24 | buffer[12]<<16 | buffer[11] << 8 | buffer[10];
  
  printf("%d\n%d\n", destination->Offset, destination->Size);
}

/**********************************************************************************************
 * This method will write a bitmap header to a given file, storing the data from memory.
 *
 * @param BitMapHeader *source This is the source of the given data.
 * @param  FILE* fp This is the name of the file to write.
 **********************************************************************************************/
 void writeBitmapHeader(BitMapHeader *source, FILE* fp)
{
 unsigned char buffer[14];

  buffer[1] = source->Type >> 8;
  buffer[0] = source->Type;
  buffer[5] = source->Size>>24;
  buffer[4] = source->Size >> 16;
  buffer[3] = source->Size >> 8;
  buffer[2] = source->Size;
  buffer[7] = source->Reserved1>>8;
  buffer[6] = source->Reserved1;
  buffer[9] = source->Reserved2>>8;
  buffer[8] = source->Reserved2;
  buffer[13] = source->Offset >> 24;
  buffer[12] = source->Offset >> 16;
  buffer[11] = source->Offset >> 8;
  buffer[10] = source->Offset;

  fwrite(&buffer, 14,1,  fp);
}


/**********************************************************************************************
 * This method will read a bitmap info structure from a given file, loading the data into memory.
 *
 * @param  BitMapHeader* destination This is where the data is to be placed.
 * @param  FILE* fp This is the file to open.
 **********************************************************************************************/
void readBitMapInfoStructure(BitMapInfo *destination , FILE* fp)
{
  unsigned char buffer[40];

  fread(&buffer, 40, 1, fp);

  destination->Size=buffer[3]<<24 | buffer[2]<<16 | buffer[1] << 8 | buffer[0];
  destination->Width=buffer[7]<<24 | buffer[6]<<16 | buffer[5] << 8 | buffer[4];
  destination->Height=buffer[11]<<24 | buffer[10]<<16 | buffer[9] << 8 | buffer[8];
  destination->Planes= buffer[13] << 8 | buffer[12];
  destination->Bits= buffer[15] << 8 | buffer[14];
  destination->Compression=buffer[19]<<24 | buffer[18]<<16 | buffer[17] << 8 | buffer[16];
  destination->ImageSize=buffer[23]<<24 | buffer[22]<<16 | buffer[21] << 8 | buffer[20];
  destination->xResolution=buffer[27]<<24 | buffer[26]<<16 | buffer[25] << 8 | buffer[24];
  destination->yResolution=buffer[31]<<24 | buffer[30]<<16 | buffer[29] << 8 | buffer[28];
  destination->Colors=buffer[35]<<24 | buffer[34]<<16 | buffer[33] << 8 | buffer[32];
  destination->ImportantColors=buffer[39]<<24 | buffer[38]<<16 | buffer[37] << 8 | buffer[36];
}

/**********************************************************************************************
 * This method will write a bitmap info structure to a given file, storing the data from memory.
 *
 * @param BitMapHeader *source This is the source of the given data.
 * @param  FILE* fp This is the name of the file to write.
 **********************************************************************************************/
void writeBitMapInfoStructure(BitMapInfo *src , FILE* fp)
{
  unsigned char buffer[40];
  buffer[0] = src->Size;
buffer[1] = src->Size >> 8;
buffer[2] = src->Size >> 16;
buffer[3] = src->Size >> 24;
buffer[4] = src->Width >> 0;
buffer[5] = src->Width >> 8;
buffer[6] = src->Width >> 16;
buffer[7] = src->Width >> 24;
buffer[8] =  src->Height >> 0;
buffer[9] = src->Height >> 8;
buffer[10] = src->Height >> 16;
buffer[11] = src->Height >> 24;

buffer[12] = src->Planes >> 0;
buffer[13] = src->Planes >> 8;
buffer[14] = src->Bits >> 0;
buffer[15] = src->Bits >> 8;
buffer[16] = src->Compression >> 0;
buffer[17] = src->Compression >> 8;
buffer[18] = src->Compression >> 16;
buffer[19] = src->Compression >> 24;
buffer[20] = src->ImageSize >> 0;
buffer[21] = src->ImageSize >> 8;
buffer[22] = src->ImageSize >> 16;
buffer[23] = src->ImageSize >> 24;
buffer[24] = src->xResolution >> 0;
buffer[25] = src->xResolution >> 8;
buffer[26] = src->xResolution >> 16;
buffer[27] = src->xResolution >> 24;
buffer[28] = src->yResolution >> 0;
buffer[29] = src->yResolution >> 8;
buffer[30] = src->yResolution >> 16;
buffer[31] = src->yResolution >> 24;
buffer[32] = src->Colors >> 0;
buffer[33] = src->Colors >> 8;
buffer[34] = src->Colors >> 16;
buffer[35] = src->Colors >> 24;
buffer[36] = src->ImportantColors >> 0;
buffer[37] = src->ImportantColors >> 8;
buffer[38] = src->ImportantColors >> 16;
buffer[39] = src->ImportantColors >> 24;

  fwrite(&buffer, 40,1,  fp);

}


BitMapHeader *createBitmapHeader(int width, int height)
{
	BitMapHeader *retVal = (BitMapHeader*)malloc(sizeof(BitMapHeader));
	retVal->Type = 0x4D42; // BM is the type, using 42 $D as the hex values.
	
    int rowSize = ((24 * width + 31) / 32)*4;
    int pixelArraySize = rowSize * height;
	
	retVal->Size = pixelArraySize + 54; 
	retVal->Reserved1 = 0;
	retVal->Reserved2 = 0;
	retVal->Offset = 54;
	
	return retVal;
}
BitMapInfo *createBitMapInfoStructure(int width, int height)
{
   BitMapInfo *retVal = (BitMapInfo*)malloc(sizeof(BitMapInfo));
   int rowSize = ((24 * width + 31) / 32)*4;
   int pixelArraySize = rowSize * height;
   
   retVal->Size=40;
   retVal->Width=width;
   retVal->Height=height;
   retVal->Planes=1;
   retVal->Bits=24;
   retVal->Compression=0;
   retVal->ImageSize=pixelArraySize;
   retVal->xResolution = 0;
   retVal->yResolution = 0;
   retVal->Colors = 0;
   retVal->ImportantColors=0;
	
   return retVal;
}








