// bitmap.h
#ifndef _BITMAP_H
#define _BITMAP_H

#include <stdio.h>

struct BitMapHeader
{
   unsigned short int Type;
   unsigned int Size;
   unsigned short int Reserved1,;
   unsigned short int Reserved2;
   unsigned int Offset;
};

struct BitMapInfo
{
   unsigned int Size;
   int Width;
   int Height;
   unsigned short int Planes;
   unsigned short int Bits;
   unsigned int Compression;
   unsigned int ImageSize;
   int xResolution;
   int yResolution;
   unsigned int Colors;
   unsigned int ImportantColors;
} ;

struct hsiPixel
{
  double hue;
  double saturation;
  double intensity;
};

struct Pixel
{
   unsigned char Green,
                 Red, 
				 Blue;
   struct hsiPixel hsirepresentation;
};

struct BitMap
{
  BitMapHeader header;
  BitMapInfo info;
  int width;
  int height;
  void* rawDataArray;
  struct Pixel *pixelArray;
  struct Pixel *rawPixels;
};

/**********************************************************************************************
 * This method will create a blank BitMap structure from headers that are passed in.
 *
 * @param  struct BitMapHeader *hdr This is the header that is to be used for this bitmap.
 * @param  struct BitMapInfo* bmi This is the bitmap information that is to eb used.
 * @return A blank bitmap will be pointed to by the return value.
 **********************************************************************************************/
struct BitMap *createBlankBitmapFromHeaders(struct BitMapHeader *hdr, struct BitMapInfo* bmi);

/**********************************************************************************************
 * This method will create a blank BitMap structure from the image height and width.
 *
 * @param  int width This is the width of the image.
 * @param  int height This is the height of the image.
 * @return A blank bitmap will be pointed to by the return value.
 **********************************************************************************************/
struct BitMap *createBlankBitmap(int width, int height);

/**********************************************************************************************
 * This function will return a pointer to a given pixel.
 *
 * @param  struct BitMap *bmp This is a pointer to the bitmap that is to be used.
 * @param  int x This is the x offset (horizontal) into the image.
 * @param  int y This is the y offset (vertical) into the image.
 * @return A pointer to the pixel will be returned.
 **********************************************************************************************/
Pixel* getPixel(struct BitMap* bmp, int x, int y);

/**********************************************************************************************
 * This function will update the internal HSI pixels from the RGB data/.
 *
 * @param  struct BitMap *bmp This is a pointer to the bitmap that is to be used.
 * @return No return from the method.
 **********************************************************************************************/
void updateHSIPixels(struct BitMap* bmp);

/**********************************************************************************************
 * This function will update the internal RGB pixels from the HSI data/.
 *
 * @param  struct BitMap *bmp This is a pointer to the bitmap that is to be used.
 * @return No return from the method.
 **********************************************************************************************/
void updateRGBPixels(struct BitMap* bmp);

/**********************************************************************************************
 * This method will destroy a given bitmap, freeing all memory.
 *
 * @param  struct BitMap* This si the bitmap to be freed.
 **********************************************************************************************/
void destroyBitmap(struct BitMap* bmp);

/**********************************************************************************************
 * This method will read a bitmap from a given file, loading the data into memory.
 *
 * @param  char* filename This si the name of the file to open.
 * @return A bitmap will be pointed to by the return value.
 **********************************************************************************************/
struct BitMap *readBitmapFromFile(char *filename);

/**********************************************************************************************
 * This method will write a bitmap to a given file, storing the data from memory.
 *
 * @param  char* filename This si the name of the file to write.
 * @param struct BitMap bitmap This si the bitmap to write out.
 **********************************************************************************************/
void writeBitmapToFile(char *filename, struct BitMap *bmp);

/**********************************************************************************************
 * This method will read a bitmap header from a given file, loading the data into memory.
 *
 * @param  BitMapHeader* destination This is where the header data is to be placed.
 * @param  FILE* fp This is the name of the file to open.
 **********************************************************************************************/
void readBitmapHeader(BitMapHeader *destination, FILE* fp);

/**********************************************************************************************
 * This method will write a bitmap header to a given file, storing the data from memory.
 *
 * @param BitMapHeader *source This is the source of the given data.
 * @param  FILE* fp This is the name of the file to write.
 **********************************************************************************************/
 void writeBitmapHeader(BitMapHeader *source, FILE* fp);

/**********************************************************************************************
 * This method will read a bitmap info structure from a given file, loading the data into memory.
 *
 * @param  BitMapHeader* destination This is where the data is to be placed.
 * @param  FILE* fp This is the file to open.
 **********************************************************************************************/
void readBitMapInfoStructure(BitMapInfo *destination , FILE* fp);

/**********************************************************************************************
 * This method will write a bitmap info structure to a given file, storing the data from memory.
 *
 * @param BitMapHeader *source This is the source of the given data.
 * @param  FILE* fp This is the name of the file to write.
 **********************************************************************************************/
void writeBitMapInfoStructure(BitMapInfo *src , FILE* fp);
BitMapHeader *createBitmapHeader(int width, int height);
BitMapInfo *createBitMapInfoStructure(int width, int height);

#endif