#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "bitmap.h"
#include "edgeDetector.h"
#include "convolutions.h"

static int gxArray[9] = {1, 0, -1, 2, 0, -2, 1, 0, -1};
static int gyArray[9] = {1, 2, 1, 0, 0, 0, -1, -2, -1};
 
/**********************************************************************************************
 * This method will perform a sobel filter on a given pixel set.
 *
 * @param Pixel *pixels These are the pixels that are to be manipulated (9).
 * @param Pixel* retValue This si the return value representing the reulting pixel.
 **********************************************************************************************/
 void SobelFilter (Pixel *pixels, Pixel* retValue)
{
  int i, j;
  int xSum = 0;
  int ySum = 0;
  
  // Convolve over the two arrays
  int gx[3];
  int gy[3];
  convolve3x3(pixels, &gxArray[0], &gx[0]);
  convolve3x3(pixels, &gyArray[0], &gy[0]);
  
  retValue->Red = sqrt(gx[0] * gx[0] + gy[0]*gy[0]);
  retValue->Green = sqrt(gx[1] * gx[1] + gy[1]*gy[1]);
  retValue->Blue = sqrt(gx[2] * gx[2] + gy[2]*gy[2]);
}


/**********************************************************************************************
 * This method will perform a sobel filter on a given pixel set.
 *
 * @param Pixel *pixels These are the pixels that are to be manipulated (9).
 * @param Pixel* retValue This si the return value representing the reulting pixel.
 **********************************************************************************************/
 void HSISobelFilter (struct hsiPixel *pixels, hsiPixel* retValue, int type)
{
  int i, j;
  int xSum = 0;
  int ySum = 0;
  
  // Convolve over the two arrays
  double gx = hsiConvolve3x3(pixels, &gxArray[0], type);
  double gy = hsiConvolve3x3(pixels, &gyArray[0], type);
  
     if (type == HUE)
	{
	  retValue->hue = sqrt(gx*gx+gy*gy);
	}
	else if (type == SATURATION)
	{
	  retValue->saturation = sqrt(gx*gx+gy*gy);
	}
	else if (type == INTENSITY)
	{
	  retValue->intensity = sqrt(gx*gx+gy*gy);
	}
}
  
/**********************************************************************************************
 * This function will detect edges in a given image.  It does so by using a sobel filter across the RGB
 * aspects of the image.  It works best with a grayscale image.
 *
 * @param struct BitMap *srcimage This si the source image that is to be modified.
 **********************************************************************************************/
struct BitMap *detectEdges(struct BitMap *srcimage)
{
    //struct BitMap *retValue = createBlankBitmapFromHeaders(&srcimage->header, &srcimage->info);
    struct BitMap *retValue = createBlankBitmap(srcimage->width, srcimage->height);
	struct Pixel pixelsToProcess[9];

    // Sweep over the image and apply a Sobel operator on the image.
    int x, y;
	int i, j;
    for(y=srcimage->height-2; y > 1 ; y--) {
      for(x=0; x < srcimage->width-2; x++) {
	    // Pixel* p = getPixel(srcimage, y, x);
	  	memcpy(&pixelsToProcess[0], getPixel(srcimage, x-1, y-1), sizeof(Pixel));
		memcpy(&pixelsToProcess[1], getPixel(srcimage, x, y-1), sizeof(Pixel));
		memcpy(&pixelsToProcess[2], getPixel(srcimage, x+1, y-1), sizeof(Pixel));
		memcpy(&pixelsToProcess[3], getPixel(srcimage, x-1, y), sizeof(Pixel));
		memcpy(&pixelsToProcess[4], getPixel(srcimage, x, y), sizeof(Pixel));
		memcpy(&pixelsToProcess[5], getPixel(srcimage, x+1, y), sizeof(Pixel));
		memcpy(&pixelsToProcess[6], getPixel(srcimage, x-1, y+1), sizeof(Pixel));
		memcpy(&pixelsToProcess[7], getPixel(srcimage, x, y+1), sizeof(Pixel));
		memcpy(&pixelsToProcess[8], getPixel(srcimage, x+1, y+1), sizeof(Pixel));
	  
        SobelFilter(&pixelsToProcess[0], getPixel(retValue, x, y));		
      }
	}  
	
	return retValue;
}

/**********************************************************************************************
 * This function will detect changes in color in a given image.  It does so by using a sobel filter across the hue
 * aspects of the image.  
 *
 * @param struct BitMap *srcimage This is the source image that is to be modified.
 **********************************************************************************************/
struct BitMap *detectColorEdges(struct BitMap *srcimage)
{
    //struct BitMap *retValue = createBlankBitmapFromHeaders(&srcimage->header, &srcimage->info);
    struct BitMap *retValue = createBlankBitmap(srcimage->width, srcimage->height);
	struct hsiPixel pixelsToProcess[9];
	
	updateHSIPixels(srcimage);

    // Sweep over the image and apply a Sobel operator on the image.
    int x, y;
	int i, j;
    for(y=srcimage->height-2; y > 1 ; y--) {
      for(x=0; x < srcimage->width-2; x++) {
	    // Pixel* p = getPixel(srcimage, y, x);
	  	memcpy(&pixelsToProcess[0], &getPixel(srcimage, x-1, y-1)->hsirepresentation, sizeof(struct hsiPixel));
	  	memcpy(&pixelsToProcess[1], &getPixel(srcimage, x-0, y-1)->hsirepresentation, sizeof(struct hsiPixel));
	  	memcpy(&pixelsToProcess[2], &getPixel(srcimage, x+1, y-1)->hsirepresentation, sizeof(struct hsiPixel));
	  	memcpy(&pixelsToProcess[3], &getPixel(srcimage, x-1, y)->hsirepresentation, sizeof(struct hsiPixel));
	  	memcpy(&pixelsToProcess[4], &getPixel(srcimage, x-0, y)->hsirepresentation, sizeof(struct hsiPixel));
	  	memcpy(&pixelsToProcess[5], &getPixel(srcimage, x+1, y)->hsirepresentation, sizeof(struct hsiPixel));
	  	memcpy(&pixelsToProcess[6], &getPixel(srcimage, x-1, y+1)->hsirepresentation, sizeof(struct hsiPixel));
	  	memcpy(&pixelsToProcess[7], &getPixel(srcimage, x-0, y+1)->hsirepresentation, sizeof(struct hsiPixel));
	  	memcpy(&pixelsToProcess[8], &getPixel(srcimage, x+1, y+1)->hsirepresentation, sizeof(struct hsiPixel));
	  
        HSISobelFilter(&pixelsToProcess[0], &getPixel(retValue, x, y)->hsirepresentation, HUE);
      }
	}  
	updateRGBPixels(retValue);
	
	return retValue;
}

/**********************************************************************************************
 * This function will detect changes in color in a given image.  It does so by using a sobel filter across the hue
 * aspects of the image.  
 *
 * @param struct BitMap *srcimage This is the source image that is to be modified.
 **********************************************************************************************/
struct BitMap *detectIntensityEdges(struct BitMap *srcimage)
{
    //struct BitMap *retValue = createBlankBitmapFromHeaders(&srcimage->header, &srcimage->info);
    struct BitMap *retValue = createBlankBitmap(srcimage->width, srcimage->height);
	struct hsiPixel pixelsToProcess[9];
	
	updateHSIPixels(srcimage);

    // Sweep over the image and apply a Sobel operator on the image.
    int x, y;
	int i, j;
    for(y=srcimage->height-2; y > 1 ; y--) {
      for(x=0; x < srcimage->width-2; x++) {
	    // Pixel* p = getPixel(srcimage, y, x);
	  	memcpy(&pixelsToProcess[0], &getPixel(srcimage, x-1, y-1)->hsirepresentation, sizeof(struct hsiPixel));
	  	memcpy(&pixelsToProcess[1], &getPixel(srcimage, x-0, y-1)->hsirepresentation, sizeof(struct hsiPixel));
	  	memcpy(&pixelsToProcess[2], &getPixel(srcimage, x+1, y-1)->hsirepresentation, sizeof(struct hsiPixel));
	  	memcpy(&pixelsToProcess[3], &getPixel(srcimage, x-1, y)->hsirepresentation, sizeof(struct hsiPixel));
	  	memcpy(&pixelsToProcess[4], &getPixel(srcimage, x-0, y)->hsirepresentation, sizeof(struct hsiPixel));
	  	memcpy(&pixelsToProcess[5], &getPixel(srcimage, x+1, y)->hsirepresentation, sizeof(struct hsiPixel));
	  	memcpy(&pixelsToProcess[6], &getPixel(srcimage, x-1, y+1)->hsirepresentation, sizeof(struct hsiPixel));
	  	memcpy(&pixelsToProcess[7], &getPixel(srcimage, x-0, y+1)->hsirepresentation, sizeof(struct hsiPixel));
	  	memcpy(&pixelsToProcess[8], &getPixel(srcimage, x+1, y+1)->hsirepresentation, sizeof(struct hsiPixel));
	  
        HSISobelFilter(&pixelsToProcess[0], &getPixel(retValue, x, y)->hsirepresentation, INTENSITY);
      }
	}  
	updateRGBPixels(retValue);
	return retValue;
}